/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 4; tab-width: 4 -*- */
/*
 * main.c
 * Copyright (C) Canonical LTD. by David Mandala 2008 <davidm@canonical.com>
 * 
 * main.c is free software.
 * 
 * You may redistribute it and/or modify it under the terms of the
 * GNU General Public License, as published by the Free Software
 * Foundation; either version 3 of the License, or (at your option)
 * any later version.
 * 
 * main.c is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with main.c.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#define OPEN_MODE  S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH
#define OPEN_FLAGS O_WRONLY | O_CREAT | O_APPEND
#define PATH_MODE 0755
#define NEW_UMASK 0

enum translation_types { dtd, properties };
enum translation_line { single, multi };
enum translated { accurate, no  };
enum jar_in_use {false, true};

const char * dtd_open = "<!ENTITY ";
const char * dtd_mid = " \"";
const char * dtd_close = "\">\n";
const char * dtd_empty_string = "";

const char * properties_equal = "=";
const char * properties_close = "\n";

#define MAX_STRING_SIZE 2000

struct translationblock {
    char * comment;
    char msgid[MAX_STRING_SIZE+1];
	int msgid_size;
    char msgstr[MAX_STRING_SIZE+1];
	int msgstr_size;
    char key[MAX_STRING_SIZE+1];
    char filename[MAX_STRING_SIZE+1];
    enum   translation_types type;
	enum   translation_line  line;
	enum   translated translation_accurate;
};

struct po_info {
	char * file_data;
	char application_name[MAX_STRING_SIZE+1];
    char open_xpi_filename[MAX_STRING_SIZE+1];
    int  fd_xpi_file;
    char language[6];
    int  type;
	enum jar_in_use jar;
    struct translationblock info;
};

/*********************************************************************
 *
 * Find the \n and make it a \0 so we can use string functions on it.
 * advance the pointer past the \0 only if we converted \n to \0.  We
 * don't want to walk off the end of the pointer.
 *
 *********************************************************************/
char * set_end_of_string(char * string, const char match)
{
	while (*string != match && *string != '\0'){
		string++;
	}
	if (*string == match){
		*string='\0';
		string++;
	}
	return(string);
}

/*********************************************************************
 *
 * See if the word fuzzy exists in the string, if so mark the 
 * translation as not accurate.
 *
 *********************************************************************/
void * check_fuzzy(char * string, struct po_info * block)
{
	char * next;
	next = set_end_of_string(string, '\n');
	if (strstr(string, "fuzzy")!=NULL){
		block->info.translation_accurate = no;
	}
	return(next);
}

/*********************************************************************
 *
 * Process a comment line.
 * #: xulrunner/de.xpi/de.jar!/locale/de/necko/necko.properties:43(3)
 *********************************************************************/
void * process_comment_line(char * string, struct po_info * block )
{
	char * start = NULL;
	char * end;
	char * next;
	
	/* Get the path and file name */
	next = set_end_of_string(string, '\n');
	while(*(++string) == ' ' && *string != '\0');
	end = start = string;
	if (start != NULL){
		if (strstr(start, ".properties:") != NULL){
			block->info.type = properties;
			end = strrchr(start, ':');
		}else if (strstr(start, ".dtd(") != NULL){
			block->info.type = dtd;
			end = strrchr(start, '(');
		} else {
			printf("WARNING: not a dtd file, nor properties (%s) \n", start);
		}
		*end = '\0';
		if (strlen(start)){
			if(strcmp(block->info.filename, start)){
				strcpy(block->info.filename, start);
			}
		}
		if (block->info.type == properties){
			*end = ':';
			end++;
		}else if (block->info.type == dtd){
			*end = '(';
		}
		/* now get the key value */
		string = end;
		start = strchr(string, '(');
		start++;
		end = strrchr(start, ')');
		*end='\0';
		if (strlen(start)){
			strcpy (block->info.key, start);
		}
		*end=')';
		end++;
		/* set some properties of the file. */
		if (strchr(block->info.filename, '!')!=NULL){
			block->jar = true;
		}
	}
	return(next);
}

/*********************************************************************
 *
 * grab all of the text data from the msg no matter if it's a msgid or
 * a msgstr they are the same, single or multiline.
 *
 *********************************************************************/
void * process_msg_data(char * string, char ** next, char * storage)
{
	int size =0;
	int more =1;
	char * start;
	char * end;

	/* First grab a single line if it's there */
	start = strchr(string, '"');
	start++;
	end = strrchr(start, '"'); *end = '\0';
	if (strlen(start)){
		strcpy(storage, start);
		*end = '"';
	}else{
		/* there was no single line present so setup for a multiline */
		char * marker;	
		*end = '"';
		end = *next;
		size = 1;
		memset(storage, 0, MAX_STRING_SIZE);
		do {
			string = end;
			end = set_end_of_string(string, '\n');
			/* test to see if we have hit msg, or a blank line or a comment line, or a null, if not get it */
			if (strncmp(string, "msg", 3) && *string != ' ' && *string != '#'&& *string != '\0' ){
				int csize = 0;
				start =  strchr(string, '"');
				start++;
				marker = strrchr(start,  '"');
				*marker = '\0';
				csize = strlen(start);
				if (csize){
					size+=csize;
					/* *storage = realloc( *storage, size);*/
					strcat(storage, start);
				}
				*marker='"';
			}else{
				more = 0;
				/* We have stepped into the next line and its not what */
				/* we want so we need to step back to the prior point */
				/* backup one place, dump in the \n that was there and */
				/* step back to the head of the line. */
				end--;
				*end = '\n';
				end=string;
			}
		}while(more);
		*next = end;
	}

	return(storage);
}

/*********************************************************************
 *
 * Process a msgid of any kind.
 *
 *********************************************************************/
void * process_msgid_line(char * string, struct po_info ** data)
{
	char * next;
	
	next = set_end_of_string(string, '\n');
	process_msg_data(string, &next, (*data)->info.msgid);
    if (((*data)->info.msgid_size = strlen((*data)->info.msgid)) == 0){
	    (*data)->info.translation_accurate = no;
	}
	return(next);
}

/*********************************************************************
 *
 * Process a msgstr of any kind.
 *
 *********************************************************************/
void * process_msgstr_line(char * string, struct po_info ** data)
{
	char * next;
	
	next = set_end_of_string(string, '\n');
	process_msg_data(string, &next, (*data)->info.msgstr);
	if(((*data)->info.msgstr_size = strlen((*data)->info.msgstr)) == 0){
		(*data)->info.translation_accurate = no;
	}
	return(next);
}

/*********************************************************************
 *
 * This function will locate and process the next translation block
 *
 *********************************************************************/
int parse_po_file_data(struct po_info * data)
{
	int ret = 1;
	char * string;
	
	
	string = data->file_data;
	/* Set translation as accurate */
	data->info.translation_accurate = accurate;
	
    while (*string++ != '#' && *string !='\0');
	if (*string == ':'){
		/* process comment line */
        string = process_comment_line(string, data);
	    if (*string == '#' && *(string+1) == ','){
			string = check_fuzzy(string, data);
		}
		if (strstr(string, "msgid")!=NULL){
			/* process what is a msgid line */
			string = process_msgid_line(strstr(string, "msgid"), &data);
		}
		if (strstr(string, "msgstr")!=NULL){
			/* process msgstr line */
			string = process_msgstr_line(strstr(string, "msgstr"), &data);
		}
		ret = 1;
	}else if (*string == '\0'){
		/* exit end of data */
		ret = -1;
	}else{
		ret=0;
	}
	data->file_data = string;
    return(ret);
}

/*********************************************************************
 *
 * verify or make a path.
 *
 *********************************************************************/
int mkfilepath(char * in_path )
{
	int ret = -1;
	char path[MAX_STRING_SIZE+1];
	char * ppath;
	char * file;
	char * end;
	/* just in case it's a const, operate on a local copy */
	if (strcpy(path, in_path)){
		ppath = path;
		/* get the file name, it's always there */
		if ((file = strrchr(path, '/'))!=NULL){
			file++;
		}
		/* Now get the path and make sure it exists or make it. */
		/* Only fail if the path does not exist and can't be made. */
		/* Step past first / if present */
		if (*ppath == '/'){
			ppath++;
		}
		end = ppath;
		do {
			end = set_end_of_string(end, '/');
			if ((ret = mkdir( path, 0755 ))== -1 && errno != EEXIST){
				break;
			}
			ret = 0;
			*(end-1)='/';
		} while (end != file);
		/*free(path);*/
	}
	return(ret);
}

char *po2dtd_value (const char *value)
{
  char* val_ptr = value;
  int len = strlen (value);
  char* buf = malloc (len * 6 * sizeof(char)+1);
  char* cur = buf;

  while(*val_ptr) {
    switch(*val_ptr) {
    case '\\': 
    {
      switch(*(val_ptr + 1)) {
      case '\\':
        *cur = '\\';
        val_ptr++; cur++;
        break;
      case 'n':
        *cur = '\n';
        val_ptr++; cur++;
        break;
      case 't':
        *cur = '\t';
        val_ptr++; cur++;
        break;
      case 'r':
        *cur = ' ';
        val_ptr++; cur++;
        break;
      case '"':
        // we replace the \ infront of quotes 
        strncpy (cur, "&quot;", 6);
        val_ptr++; cur+=6;
        break;
      }
      break;
    }
    case '%':
      strncpy (cur, "&#037;", 6);
      cur+=6;
      break;
    default:
      *cur = *val_ptr;
      cur++;
      break;
    }
    val_ptr++;
  }
  *cur = 0;

  return buf;
}

int write_encoded(int fd, const char* value, int length)
{
  char *value_enc = po2dtd_value (value);
  write (fd, value_enc, strlen(value_enc));
  free(value_enc);
  return length;
}

/*********************************************************************
 *
 * write out the data in the dtd format:
 * <!ENTITY key data >\n
 * which could be:
 * When we have a translation:
 * dtd_open data->info.key data->info.msgstr dtd_close
 * When we don't have a translation but do have an msgid
 * dtd_open data->info.key data->info.msgid dtd_close
 * When we don't even have a msgid
 * dtd_open data->info.key dtd_empty_string dtd_close
 *
 *********************************************************************/
int write_dtd_format_data(struct po_info * data)
{
	int length;
	int count;

	length = strlen(dtd_open);
	count = write(data->fd_xpi_file, dtd_open, length);
	if (length == count){
		length = strlen(data->info.key);
		count = write(data->fd_xpi_file, data->info.key, length);
		if (length == count){
			if ( data->info.translation_accurate == accurate){
				length = strlen(dtd_mid);
				count = write(data->fd_xpi_file, dtd_mid, length);
				if (length != count) {
					goto exit_now;
				}
				if (data->info.msgstr_size &&
				    strcmp (data->info.key, data->info.msgstr)) {
					length = strlen(data->info.msgstr);
					count = write_encoded(data->fd_xpi_file, data->info.msgstr, length);
					if (length != count){
						goto exit_now;
					}
				} else {
					length = strlen(dtd_empty_string);
					count = write(data->fd_xpi_file, dtd_empty_string,length);
					if (length != count){
						goto exit_now;
					}
				}
			} else {
				length = strlen(dtd_mid);
				count = write(data->fd_xpi_file, dtd_mid, length);
				if (length != count){
					goto exit_now;
				}
				if (data->info.msgid_size &&
				    strcmp (data->info.key, data->info.msgid)) {
					length = strlen(data->info.msgid);
					count = write_encoded(data->fd_xpi_file, data->info.msgid, length);
					if (length != count){
						goto exit_now;
					}
				} else {
					length = strlen(dtd_empty_string);
					count = write(data->fd_xpi_file, dtd_empty_string,length);
					if (length != count){
						goto exit_now;
					}
				}
			}
			length = strlen(dtd_close);
			count = write(data->fd_xpi_file, dtd_close, length);
			if (length == count){
				count = 0;
			}
		}
	}
	exit_now:
	return(count);
}

/*********************************************************************
 *
 * write out the data into the properties format.
 * application.translation.key data\n
 * which could be:
 * When we have a translation:
 * data->application_name .translation. data->info.key data->info.msgstr\n
 * When we don't have a translation but do have an msgid
 * data->application_name .translation. data->info.key data->info.msgid\n
 * When we don't even have a msgid
 * data->application_name .translation. data->info.key\n
 *
 *********************************************************************/
int write_properties_format_data(struct po_info * data)
{
	int length;
	int count;
	
	length = strlen(data->info.key);
	count = write(data->fd_xpi_file, data->info.key, length);
	if (length == count){
		length = strlen(properties_equal);
		count = write(data->fd_xpi_file, properties_equal, length);
		if (length == count){
			if (data->info.translation_accurate == accurate &&
			    data->info.msgstr_size &&
		    	    strcmp (data->info.key, data->info.msgstr)) {
				length = strlen(data->info.msgstr);
				count = write(data->fd_xpi_file, data->info.msgstr, length);
				if (length != count){
					goto exit_now;
				}
			}else{
				if (data->info.msgid_size &&
		    	    	    strcmp (data->info.key, data->info.msgid)) {
					length = strlen(data->info.msgid);
					count = write(data->fd_xpi_file, data->info.msgid, length);
					if (length != count){
						goto exit_now;
					}
				}
			}
			length = strlen(properties_close);
			count = write(data->fd_xpi_file, properties_close, length);
			if (length == count){
				count = 0;
			}
		}
	}
	exit_now:		
		return(count);
}

/*********************************************************************
 *
 * Write out the parsed data into the correct format.
 *
 *********************************************************************/
int write_xpi_data(struct po_info * data)
{
	int x = 0;
	int mode;
	int old_mask;
	int fd_xpi_file;
	char * filename;

	filename = data->info.filename;
	fd_xpi_file = data->fd_xpi_file;
	mode = PATH_MODE;
	old_mask = umask(NEW_UMASK);
	if (!strlen(data->open_xpi_filename)){
		if (!(x = mkfilepath(filename))){
			if ((data->fd_xpi_file = open(data->info.filename, OPEN_FLAGS, OPEN_MODE))!= -1){
				strcpy(data->open_xpi_filename, data->info.filename);
			}
		}
	}else if (strcmp(data->open_xpi_filename, data->info.filename)){
		close(data->fd_xpi_file);
		mode = PATH_MODE;
		old_mask = umask(NEW_UMASK);
		if (!(x = mkfilepath(filename))){
			if ((data->fd_xpi_file = open(filename, OPEN_FLAGS, OPEN_MODE))!= -1){
				strcpy(data->open_xpi_filename, data->info.filename);
			}
		}
	}
	umask(old_mask);

	if (data->info.type == dtd){
		write_dtd_format_data(data);
	}else if (data->info.type == properties){
		write_properties_format_data(data);
	}
	return(x);
}

void help()
{
	printf("\n\npo2xpi Version 0.1 development version\n\n");
	printf("You must supply the translate applicatation name and the data file name\n");
	printf("on the command line -> \"po2xpi xulrunner new_de.po\"\n\n");
	printf("This version of the software is pretty stupid and expect things in a\n");
	printf("specific order.  And it's going to make it's path relative to the \n");
	printf("directory it's started in\n");
}


int main(int argc, char * argv[] )
{
	int x;
	int fd;
	static struct po_info block;
	size_t size;
	char * string = NULL;
	char * data_file;
	struct po_info * pblock;

	memset(&block, 0, sizeof(struct po_info));
	/* Inbound info */
	if (argc == 3){
		strcpy(block.application_name, argv[1]);
		data_file = strdup(argv[2]);
	}else{
		help();
		exit(-1);
	}
	if ((fd = open(data_file, O_RDONLY | O_APPEND)) != -1){
		size = lseek(fd, 0, SEEK_END) + 10;
		lseek(fd, 0, SEEK_SET);
		string = block.file_data = calloc(size, sizeof(char));
		read(fd, block.file_data, size);
		close(fd);
	} else {
		printf("ERROR: couldn't open file (%s)", data_file);
	}
	while( (x = parse_po_file_data(&block)) != -1){
		if (x){
			pblock = &block;
			x = write_xpi_data( pblock );
		}
	}
	close(block.fd_xpi_file);
	free(string);
	free(data_file);
	return (0);
}

